#-----------------------------------------------------------------------------------
# Basic interlocking and route setting system tests
#
# When run as 'main' it uses the following example schematic files:
#     "../configuration_examples/interlocking_colour_light_example.sig"
#     "../configuration_examples/interlocking_semaphore_example.sig"
#
# common functions are also reused for the automation and approach control tests
#-----------------------------------------------------------------------------------

from system_test_harness import *
import test_object_edit_windows

#-----------------------------------------------------------------------------------

def run_initial_state_tests(semaphore=False):
    print("Colour Light Initial state tests")
    # Signal 14 has no valid route available
    assert_signals_route_MAIN(1,3,4,5,6,7,8,9,10,11,12,13,15,16,17,18,19)
    assert_signals_route_LH1(2)
    assert_signals_DANGER(1,2,3,4,5,6,7,8,10,11,12,13,14,15,17,19)
    assert_signals_CAUTION(9,18)
    # Differences are due to semaphore associated signals
    if semaphore:
        assert_signals_DANGER(16)
        assert_signals_locked(1016)
        assert_signals_unlocked(16)
    else:
        assert_signals_CAUTION(16)
        assert_signals_locked(16)
    assert_points_unlocked(2,3,5)
    assert_signals_unlocked(1,3,4,5,6,7,8,10,11,12,13,15,17,19)
    assert_subsidaries_unlocked(1,2,3)
    assert_signals_locked(2,14)
    # distant signals 9,16 and 18 are locked on the home signals ahead
    assert_signals_locked(9,18)
    return()

#-----------------------------------------------------------------------------------

def run_signal_route_tests():
    print("Common Signal Route tests")
    # Test the default route states
    # There is no route for signal 14 in the 'default' state
    assert_signals_route_MAIN(1,5,6,7,13)
    assert_signals_route_LH1(2)
    # Signals 1 and 6
    set_fpls_off(2)
    set_points_switched(2)
    set_fpls_on(2) 
    assert_signals_route_LH1(1,6)
    set_fpls_off(2)
    set_points_normal(2)
    set_fpls_on(2)
    assert_signals_route_MAIN(1,6)
    # Signal 2, Signal 13 and Signal 14
    # Note that for Signal 2, MAIN is for the main line and LH1 is the siding
    # Note that for Signal 14, MAIN is for the crossover (no route back along main line)
    assert_signals_route_LH1(2)
    assert_signals_route_MAIN(13)
    # There is "no route" for signal 14
    set_fpls_off(3)
    set_points_switched(3)
    set_fpls_on(3)
    assert_signals_route_MAIN(2)
    assert_signals_route_RH1(13)
    # There is "no route" for signal 14
    set_points_switched(5)
    assert_signals_route_LH1(13)
    assert_signals_route_MAIN(14)
    set_points_normal(5)
    assert_signals_route_MAIN(2)
    assert_signals_route_RH1(13)
    # There is "no route" for signal 14
    set_fpls_off(3)
    set_points_normal(3)
    set_fpls_on(3)
    # Test everything has returned to the default state
    # There is "no route" for signal 14
    assert_signals_route_MAIN(1,5,6,7,13)
    assert_signals_route_LH1(2)
    return()

#-----------------------------------------------------------------------------------

def run_signal_aspect_tests(semaphore=False):
    print("Signal Aspect tests")
    # The main route tests are different because of associated distants
    # Set the home signals 'off' first, followed by the distant signals
    if semaphore:
        # Test the default state
        assert_signals_DANGER(17,16,8,1,2,3,4)
        # Main Route 1
        set_signals_off(17,16)
        assert_signals_PROCEED(17)
        assert_signals_CAUTION(16)
        set_signals_off(8,1,3,4,1016)
        assert_signals_PROCEED(8,1,3,4,16)
        set_signals_on(16,17,8,1,3,4,1016)
        # Main Route 2 - loop
        set_fpls_off(2,3)
        set_points_switched(2,3)
        set_fpls_on(2,3)
        set_signals_off(17,16)
        assert_signals_PROCEED(17)
        assert_signals_CAUTION(16)
        set_signals_off(8,1,2,4,1016)
        assert_signals_PROCEED(8,1,2,4,16)
        set_signals_on(16,17,8,1,2,4,1016)
        set_fpls_off(2,3)
        set_points_normal(2,3)
        set_fpls_on(2,3)
        # Test everything has returned to the default state
        assert_signals_DANGER(17,16,8,1,2,3,4)
    else:
        # Test the default state
        assert_signals_DANGER(17,8,1,2,3,4)
        assert_signals_CAUTION(16)
        # Main Route 1 - block 1
        set_signals_off(8,1,3,4,16,17)
        assert_signals_PROCEED(8,1,3,4,16,17)
        set_signals_on(8,1,3,4,16,17)
        # Main Route 2 - loop
        set_fpls_off(2,3)
        set_points_switched(2,3)
        set_fpls_on(2,3)
        set_signals_off(8,1,2,4,16,17)
        assert_signals_PROCEED(8,1,2,4,16,17)
        set_signals_on(8,1,2,4,16,17)
        set_fpls_off(2,3)
        set_points_normal(2,3)
        set_fpls_on(2,3)
        # Test everything has returned to the default state
        assert_signals_DANGER(17,8,1,2,3,4)
        assert_signals_CAUTION(16)
    # Main Route 2
    # Test the default state
    assert_signals_DANGER(12,10,19,11)
    assert_signals_CAUTION(9,18)
    # Set the home signals 'off' first, followed by the distant signals
    set_signals_off(11,19,18,10,12,9)
    assert_signals_PROCEED(11,19,18,10,12,9)
    set_signals_on(11,19,18,10,12,9)
    # Test everything has returned to the default state
    assert_signals_DANGER(12,10,19,11)
    assert_signals_CAUTION(9,18)
    return()

#-----------------------------------------------------------------------------------

def run_point_interlocking_tests():
    print("Common Point Interlocking Tests")
    # Test the default state
    assert_points_unlocked(2,3,5)
    assert_signals_unlocked(1,3,5,6,7,12,13,15)
    assert_signals_locked(2,14)
    assert_subsidaries_unlocked(1,2,3)
    # Point 2 - Test interlocking when normal
    set_fpls_off(2)
    assert_signals_locked(1,5,6,15)
    assert_subsidaries_locked(1)
    set_fpls_on(2)
    assert_signals_unlocked(1,5,6,15)
    assert_subsidaries_unlocked(1)
    # Point 2 - Test interlocking when switched
    set_fpls_off(2)
    set_points_switched(2)
    set_fpls_on(2)
    assert_signals_unlocked(1,6)
    assert_signals_locked(5,15)
    assert_subsidaries_unlocked(1)
    # Point 2 - Test interlocking when returned to normal
    set_fpls_off(2)
    assert_signals_locked(1,5,6,15)
    assert_subsidaries_locked(1)
    set_points_normal(2)
    set_fpls_on(2)
    assert_signals_unlocked(1,5,6,15)
    assert_subsidaries_unlocked(1)
    # Point 3 - Test interlocking when normal
    assert_signals_unlocked(3,7,13)
    assert_subsidaries_unlocked(2,3)
    assert_signals_locked(2)
    set_fpls_off(3)
    assert_signals_locked(2,3,7,13)
    assert_subsidaries_locked(2,3)
    set_fpls_on(3)
    assert_signals_unlocked(3,7,13)
    assert_subsidaries_unlocked(2,3)
    assert_signals_locked(2)
    # Test effect of Point 5 (different signals interlocked)
    assert_signals_locked(14)
    assert_signals_unlocked(12,3,13)
    assert_subsidaries_unlocked(3)
    set_points_switched(5)
    assert_signals_locked(3,12)
    assert_signals_unlocked(13,14)
    assert_subsidaries_locked(3)
    set_points_normal(5)
    assert_signals_locked(14)
    assert_signals_unlocked(12,3,13)
    assert_subsidaries_unlocked(3)
    # Point 3 - Test interlocking when switched
    set_fpls_off(3)
    set_points_switched(3)
    set_fpls_on(3)
    assert_signals_unlocked(2,13)
    assert_signals_locked(3,7)
    assert_subsidaries_unlocked(2)
    assert_subsidaries_locked(3)
    # Test effect of Point 5 (different signals interlocked)
    assert_signals_locked(14)
    assert_signals_unlocked(12,2,13)
    assert_subsidaries_unlocked(2)
    set_points_switched(5)
    assert_signals_locked(2,12)
    assert_signals_unlocked(13,14)
    assert_subsidaries_locked(2)
    set_points_normal(5)
    assert_signals_locked(14)
    assert_signals_unlocked(12,2,13)
    assert_subsidaries_unlocked(2)
    # Point 2 - Test interlocking when returned to normal
    set_fpls_off(3)
    assert_signals_locked(2,3,7,13)
    assert_subsidaries_locked(2,3)
    set_points_normal(3)
    set_fpls_on(3)
    assert_signals_unlocked(3,7,13)
    assert_signals_locked(2)
    assert_subsidaries_unlocked(2,3)
    # Point 3 - Test interlocking when normal
    assert_signals_unlocked(3,13,12)
    assert_signals_locked(14)
    assert_subsidaries_unlocked(3)
    # Point 3 - Test interlocking when switched
    set_points_switched(5)
    assert_signals_unlocked(13,14)
    assert_signals_locked(3,12)
    assert_subsidaries_locked(3)
    # Point 2 - Test interlocking when returned to normal
    set_points_normal(5)
    # Test everything has been returned to the default state
    assert_points_unlocked(2,3,5)
    assert_signals_unlocked(1,3,5,6,7,12,13,15)
    assert_signals_locked(2,14)
    assert_subsidaries_unlocked(1,2,3)
    return()

#-----------------------------------------------------------------------------------

def run_signal_interlocking_tests():
    print("Common Signal Interlocking Tests")
    # Test the default state
    assert_points_unlocked(2,3,5)
    assert_signals_unlocked(1,3,5,6,7,12,13,15)
    assert_signals_locked(2,14)
    assert_subsidaries_unlocked(1,2,3)
    # Test signal 8 interlocking
    set_signals_off(8)
    assert_signals_locked(15)
    set_signals_on(8)
    assert_signals_unlocked(15)
    # Test effect of point 2 (different signals interlocked)
    set_fpls_off(2)
    set_points_switched(2)
    set_fpls_on(2)
    set_signals_off(8)
    assert_signals_locked(6)
    set_signals_on(8)
    assert_signals_unlocked(6)
    # Return point 2 to normal
    set_fpls_off(2)
    set_points_normal(2)
    set_fpls_on(2)
    # Test signal 1 interlocking - main line
    set_signals_off(1)
    assert_signals_locked(13,15)
    assert_subsidaries_locked(1)
    assert_points_locked(2)
    set_signals_on(1)
    assert_signals_unlocked(13,15)
    assert_subsidaries_unlocked(1)
    assert_points_unlocked(1)
    set_subsidaries_off(1)
    assert_signals_locked(1,13,15)
    assert_points_locked(2)
    set_subsidaries_on(1)
    assert_signals_unlocked(1,13,15)
    assert_points_unlocked(1)
    # Test signals 1 interlocking on loop line
    set_fpls_off(2)
    set_points_switched(2)
    set_fpls_on(2)
    set_signals_off(1)
    assert_signals_locked(6,7)
    assert_points_locked(2)
    set_signals_on(1)
    set_subsidaries_off(1)
    assert_signals_locked(6,7)
    assert_points_locked(2)
    set_subsidaries_on(1)
    assert_signals_unlocked(6,7)
    assert_points_unlocked(1)
    # Test effect of point 3 (different interlocked signals)
    set_fpls_off(3)
    set_points_switched(3)
    set_fpls_on(3)
    set_signals_off(1)
    assert_signals_locked(6,13)
    set_signals_on(1)
    assert_signals_unlocked(6,13)
    set_subsidaries_off(1)
    assert_signals_locked(6,13)
    set_subsidaries_on(1)
    assert_signals_unlocked(6,13)
    assert_points_unlocked(1)
    # Set Points 2 and 3 back to their default states
    set_fpls_off(3)
    set_points_normal(3)
    set_fpls_on(3)
    set_fpls_off(2)
    set_points_normal(2)
    set_fpls_on(2)
    # Test signal 15 interlocking
    set_signals_off(15)
    assert_signals_locked(1,8)
    assert_subsidaries_locked(1)
    assert_points_locked(2)
    set_signals_on(15)
    assert_signals_unlocked(1,8)
    assert_subsidaries_unlocked(1)
    assert_points_unlocked(2)
    # Test Signal 3 interlocking
    set_signals_off(3)
    assert_signals_locked(13)
    assert_subsidaries_locked(3)
    assert_points_locked(3,5)
    set_signals_on(3)
    assert_signals_unlocked(13)
    assert_subsidaries_unlocked(3)
    assert_points_unlocked(2)
    set_subsidaries_off(3)
    assert_signals_locked(13,3)
    assert_points_locked(3,5)
    set_subsidaries_on(3)
    assert_signals_unlocked(13,3)
    assert_points_unlocked(3,5)
    # Test Signal 13 interlocking for MAIN route
    set_signals_off(13)
    assert_signals_locked(1,3)
    assert_subsidaries_locked(1,3)
    assert_points_locked(3,5)
    set_signals_on(13)
    assert_signals_unlocked(1,3)
    assert_subsidaries_unlocked(1,3)
    assert_points_unlocked(3,5)
    # Test Signal 13 interlocking for LH1 route
    set_points_switched(5)
    set_signals_off(13)
    assert_signals_locked(14)
    assert_points_locked(5)
    set_signals_on(13)
    assert_signals_unlocked(14)
    assert_points_unlocked(5)
    set_points_normal(5)
    # Test Signal 13 interlocking for RH1 route
    set_fpls_off(3)
    set_points_switched(3)
    set_fpls_on(3)
    set_signals_off(13)
    assert_signals_locked(2,5)
    assert_subsidaries_locked(2)
    assert_points_locked(3,5)
    set_signals_on(13)
    assert_signals_unlocked(2,5)
    assert_subsidaries_unlocked(2)
    assert_points_unlocked(3,5)
    # Test effect of point 2 (different interlocked signals)
    set_fpls_off(2)
    set_points_switched(2)
    set_fpls_on(2)
    set_signals_off(13)
    assert_signals_locked(2,1)
    assert_subsidaries_locked(2,1)
    assert_points_locked(3,5)
    set_signals_on(13)
    assert_signals_unlocked(2,1)
    assert_subsidaries_unlocked(2,1)
    assert_points_unlocked(3,5)
    # Set Points 2 and 3 back to their default states
    set_fpls_off(3)
    set_points_normal(3)
    set_fpls_on(3)
    set_fpls_off(2)
    set_points_normal(2)
    set_fpls_on(2)
    # Test Signal 5 interlocking
    set_signals_off(5)
    assert_signals_locked(6,7)
    assert_points_locked(2)
    set_signals_on(5)
    assert_signals_unlocked(6,7)
    assert_points_unlocked(2)
    # Test effect of point 3 (different interlocked signals)
    set_fpls_off(3)
    set_points_switched(3)
    set_fpls_on(3)
    set_signals_off(5)
    assert_signals_locked(6,13)
    assert_points_locked(2)
    set_signals_on(5)
    assert_signals_unlocked(6,13)
    assert_points_unlocked(2)
    # Set Point 3 back to its default state
    set_fpls_off(3)
    set_points_normal(3)
    set_fpls_on(3)
    # Test Signal 6 interlocking
    set_signals_off(6)
    assert_signals_locked(5)
    assert_points_locked(2)
    set_signals_on(6)
    assert_signals_unlocked(5)
    assert_points_unlocked(2)
    # Test effect of point 2 (different interlocked signals)
    set_fpls_off(2)
    set_points_switched(2)
    set_fpls_on(2)
    set_signals_off(6)
    assert_signals_locked(1,8)
    assert_subsidaries_locked(1)
    assert_points_locked(2)
    set_signals_on(6)
    assert_signals_unlocked(1,8)
    assert_subsidaries_unlocked(1)
    assert_points_unlocked(2)
    # Set Point 2 back to its default state
    set_fpls_off(2)
    set_points_normal(2)
    set_fpls_on(2)
    # Test signal 7 interlocking
    # Note that signal 2 is always locked when point 3 is normal
    set_signals_off(7)
    assert_signals_locked(2,5)
    assert_subsidaries_locked(2)
    assert_points_locked(3)
    set_signals_on(7)
    assert_signals_locked(2)
    assert_signals_unlocked(5)
    assert_subsidaries_unlocked(1)
    assert_points_unlocked(3)
    # Test effect of point 2 (different interlocked signals)
    # Note that signal 2 should always be locked unless route is set for MAIN
    set_fpls_off(2)
    set_points_switched(2)
    set_fpls_on(2)
    set_signals_off(7)
    assert_signals_locked(2,1)
    assert_subsidaries_locked(2,1)
    assert_points_locked(3)
    set_signals_on(7)
    assert_signals_locked(2)
    assert_signals_unlocked(1)
    assert_subsidaries_unlocked(2,1)
    assert_points_unlocked(3)
    # Set Point 2 back to its default state
    set_fpls_off(2)
    set_points_normal(2)
    set_fpls_on(2)
    # Test Signal 2 interlocking
    # Note that main signal is only for the MAIN route (not the siding)
    assert_signals_locked(2)
    set_subsidaries_off(2)
    assert_signals_locked(7)
    assert_points_locked(3)
    set_subsidaries_on(2)
    assert_signals_unlocked(7)
    assert_points_unlocked(3)
    # Test Signal 2 interlocking for LH1 route
    set_fpls_off(3)
    set_points_switched(3)
    set_fpls_on(3)
    set_signals_off(2)
    assert_signals_locked(13)
    assert_points_locked(3)
    set_signals_on(2)
    assert_signals_unlocked(13)
    assert_points_unlocked(3)
    set_subsidaries_off(2)
    assert_signals_locked(13)
    assert_points_locked(3)
    set_subsidaries_on(2)
    assert_signals_unlocked(13)
    assert_points_unlocked(3)
    # Set Point 3 back to its default state
    set_fpls_off(3)
    set_points_normal(3)
    set_fpls_on(3)
    # Test Signal 12 interlocking
    # Note that signal 14 should always be locked for this test
    set_signals_off(12)
    assert_signals_locked(14)
    assert_points_locked(5)
    set_signals_on(12)
    assert_signals_locked(14)
    assert_points_unlocked(5)
    # Test Signal 14 interlocking
    # Note this can only be cleared for the crossover
    assert_signals_locked(14)
    set_points_switched(5)
    set_signals_off(14)
    assert_signals_locked(13)
    assert_points_locked(5)
    set_signals_on(14)
    assert_signals_unlocked(13)
    assert_points_unlocked(5)
    set_points_normal(5)
    # Test everything has been returned to the default state
    assert_points_unlocked(2,3,5)
    assert_signals_unlocked(1,3,5,6,7,12,13,15)
    assert_signals_locked(2,14)
    assert_subsidaries_unlocked(1,2,3)
    return()

#-----------------------------------------------------------------------------------

def run_signal_interlock_ahead_tests(semaphore=False):
    print("Signal Interlock on signals ahead Tests")
    if semaphore: sig_id = 1016
    else: sig_id = 16
    # Test the default state
    assert_signals_locked(9,18,sig_id)
    # Main line 1
    assert_signals_locked(sig_id)
    set_signals_off(8)
    assert_signals_locked(sig_id)
    set_signals_off(1)
    assert_signals_locked(sig_id)
    set_signals_off(3)
    assert_signals_locked(sig_id)
    set_signals_off(4)
    assert_signals_unlocked(sig_id)
    set_signals_off(sig_id)
    set_signals_on(8,1,3,4)
    assert_signals_unlocked(sig_id)
    set_signals_on(sig_id)
    assert_signals_locked(sig_id)
    # Main Line Loop
    set_fpls_off(2,3)
    set_points_switched(2,3)
    set_fpls_on(2,3)
    assert_signals_locked(sig_id)
    set_signals_off(8)
    assert_signals_locked(sig_id)
    set_signals_off(1)
    assert_signals_locked(sig_id)
    set_signals_off(2)
    assert_signals_locked(sig_id)
    set_signals_off(4)
    assert_signals_unlocked(sig_id)
    set_signals_off(sig_id)
    set_signals_on(8,1,2,4)
    assert_signals_unlocked(sig_id)
    set_signals_on(sig_id)
    assert_signals_locked()
    set_fpls_off(2,3)
    set_points_normal(2,3)
    set_fpls_on(2,3)
    # Main Line 2 - first section
    assert_signals_locked(9)
    set_signals_off(10)
    assert_signals_locked(9)
    set_signals_off(12)
    assert_signals_unlocked(9)
    set_signals_off(9)
    set_signals_on(12,10)
    assert_signals_unlocked(9)
    set_signals_on(9)
    assert_signals_locked(9)    
    # Main Line 2 - second section
    assert_signals_locked(18)
    set_signals_off(19)
    assert_signals_locked(18)
    set_signals_off(11)
    assert_signals_unlocked(18)
    set_signals_off(18)
    set_signals_on(19,11)
    assert_signals_unlocked(18)
    set_signals_on(18)
    assert_signals_locked(18)    
    # Test the default state
    assert_signals_locked(9,18,sig_id)
    return()

######################################################################################################

def run_all_interlocking_example_tests(delay:float=0.0):
    initialise_test_harness(filename="../configuration_examples/interlocking_colour_light_example.sig")
    # Edit/save all schematic objects to give confidence that editing doesn't break the layout configuration
    set_edit_mode()
    test_object_edit_windows.test_all_object_edit_windows(delay)
    set_run_mode()
    reset_layout()
    run_initial_state_tests()
    run_signal_route_tests()
    run_signal_aspect_tests()
    run_point_interlocking_tests()
    run_signal_interlocking_tests()
    run_signal_interlock_ahead_tests()
    initialise_test_harness(filename="../configuration_examples/interlocking_semaphore_example.sig")
    # Edit/save all schematic objects to give confidence that editing doesn't break the layout configuration
    set_edit_mode()
    test_object_edit_windows.test_all_object_edit_windows(delay)
    set_run_mode()
    reset_layout()
    run_initial_state_tests(semaphore=True)
    run_signal_route_tests()
    run_signal_aspect_tests(semaphore=True)
    run_point_interlocking_tests()
    run_signal_interlocking_tests()
    run_signal_interlock_ahead_tests(semaphore=True)
    report_results()
    
if __name__ == "__main__":
    start_application(lambda:run_all_interlocking_example_tests(delay=0.0))

######################################################################################################
